import json
import logging
from copy import copy
from pprint import pformat
from time import sleep

import requests
from six import wraps
from six.moves.urllib.parse import urljoin
from six.moves.urllib.parse import urlparse

logger = logging.getLogger(__name__)
MAX_RETRIES = 5
DEFAULT_TIMEOUT = 180


class ProviderHTTPClientException(Exception):
    def __init__(self, *args, **kwargs):
        super(ProviderHTTPClientException, self).__init__(*args)
        for key, value in kwargs.items():
            setattr(self, key, value)


class RetryableException(ProviderHTTPClientException):
    pass


class ProviderHTTPServiceException(ProviderHTTPClientException):
    pass


def flatten_nested_object(obj, parent_key=''):
    result = {}

    for key, value in obj.items():
        if parent_key:
            new_key = '.'.join([parent_key, key])
        else:
            new_key = key

        if isinstance(value, list):
            raise ValueError('Nesting list is not supported yet. Please change key %s' % new_key)
        if isinstance(value, dict):
            result.update(flatten_nested_object(value, new_key))
        else:
            result[new_key] = value
    return result


class ProviderRequestProxy(object):
    def __init__(self, url, headers, timeout, retries=MAX_RETRIES, host=None, retry_delay=5, **kwargs):
        self.url = url
        self.headers = headers
        self.timeout = timeout
        self.retries = retries
        self.host = host
        self.raw = kwargs.get('raw', False)
        self.is_json = kwargs.get('json', True)
        self.retry_delay = retry_delay

    def augment_request_arguments(self, **request_kwargs):
        self.headers.update(request_kwargs.get('headers', {}))
        url_override = request_kwargs.get('url', None)
        self.url = self.url.format(**request_kwargs)

        data = {
            'url': url_override if url_override else self.url,
            'headers': self.headers,
            'timeout': self.timeout,
        }
        if 'data' in request_kwargs:
            data['data'] = request_kwargs['data']
        if 'json' in request_kwargs:
            data['json'] = request_kwargs['json']
        if 'files' in request_kwargs:
            data['files'] = request_kwargs['files']
        if 'stream' in request_kwargs:
            data['stream'] = request_kwargs['stream']
        if 'params' in request_kwargs:
            data['params'] = request_kwargs['params']

        if 'json' in data and 'data' in data:
            raise ValueError('You need to use either `json` or `data` parameter, not both.')

        if ('json' in data or 'data' in data) and 'files' in data:
            # parse json or data to enable multipart
            data_payload = data.pop('json', data.pop('data', None))

            if isinstance(data_payload, str):
                data_payload = json.load(data_payload)

            if isinstance(data_payload, list):
                raise ValueError('List json is not supported with Multipart')

            data['data'] = flatten_nested_object(data_payload)

        return data

    def get(self, **request_arguments):
        return self.request_with_retry('get', **request_arguments)

    def head(self, **request_arguments):
        return self.request_with_retry('head', **request_arguments)

    def options(self, **request_arguments):
        return self.request_with_retry('options', **request_arguments)

    def post(self, **request_arguments):
        return self.request_with_retry('post', **request_arguments)

    def put(self, **request_arguments):
        return self.request_with_retry('put', **request_arguments)

    def patch(self, **request_arguments):
        return self.request_with_retry('patch', **request_arguments)

    def delete(self, **request_arguments):
        return self.request_with_retry('delete', **request_arguments)

    def format_return_data(self, response):
        if response.status_code == 204 or self.raw:
            return response
        elif self.is_json:
            try:
                return response.json()
            except ValueError as e:
                if 'JSON' not in getattr(e, 'message', ''):
                    raise
                logger.exception('Failed to decode json file with error message: %s', e)
                return {}
        else:
            return response.content

    def request_with_retry(self, method, **request_arguments):
        request_arguments = self.augment_request_arguments(**request_arguments)
        retries = 0
        retry_errors = []

        while retries < self.retries:
            try:
                response = getattr(requests, method)(**request_arguments)
                if 400 <= response.status_code < 500:
                    http_error_msg = '%s Client Error: %s for url: %s %s' % \
                                     (response.status_code,
                                      response.reason,
                                      response.url,
                                      response.content)
                    raise ProviderHTTPClientException(http_error_msg)
                # TODO: add more retryable exceptions code
                elif response.status_code in [502, 503]:
                    raise RetryableException('Deployment might be in progress, or server is overloaded')
                elif 500 <= response.status_code < 600:
                    raise ProviderHTTPClientException('%d Server error. %s ' % (
                        response.status_code,
                        response.reason
                    ), response=response)

            except (requests.Timeout, RetryableException) as exc:
                logger.debug('Retry request %s %s', method, request_arguments['url'])
                retries += 1
                retry_errors.append(exc)
                sleep(self.retry_delay)
            except requests.ConnectionError as exc:
                retry_errors.append(exc)
                raise ProviderHTTPServiceException('Service is not working with error: %s' % exc)
            else:
                break  # Exits while loop
        else:
            raise ProviderHTTPServiceException('Failed to do request after {} tries'.format(MAX_RETRIES),
                                               retries=retry_errors)

        return self.format_return_data(response)


def remote_service(path, headers=None, timeout=None, json=True, raw=False):
    if headers is None:
        headers = {}
    assert isinstance(headers, dict), 'Headers should be a dict!'

    # safe_methods = ['get', 'head', 'options']
    # unsafe_methods = ['post', 'put', 'patch', 'delete']

    def wrapper(func):
        @wraps(func)
        def inner(self, *args, **kwargs):
            if not isinstance(self, ReboticsBaseProvider):
                raise TypeError('This decorator should be used only for subclasses Rebotics Base Providers')

            provider = copy(self)

            authentication_headers = provider.get_provider_headers()
            headers.update(authentication_headers)

            session = ProviderRequestProxy(provider.build_url(path),
                                           headers,
                                           timeout=timeout or provider.timeout,
                                           retries=provider.retries,
                                           json=json,
                                           raw=raw,
                                           host=provider.host)

            # FIXME: this fucking shit is not thread safe!!!!
            provider.session = session
            self.session = session
            return func(provider, *args, **kwargs)

        return inner

    return wrapper


# noinspection PyMethodMayBeStatic
class ReboticsBaseProvider(object):
    retries = MAX_RETRIES
    timeout = DEFAULT_TIMEOUT

    def __init__(self, host, **kwargs):
        self.host = host
        self.domain = urlparse(self.host).netloc
        # TODO: add host checking

        self.headers = kwargs.get('headers', {})
        self.data = kwargs.get('data', {})
        self.session = None
        self.requests = requests

        if 'retries' in kwargs:
            self.retries = kwargs['retries']

        if 'timeout' in kwargs:
            self.timeout = kwargs['timeout']

        if 'token' in kwargs:
            self.set_token(kwargs['token'])

        self.role = kwargs.get('role', '')

    def get_provider_headers(self):
        return self.headers if self.headers else {}

    def set_header(self, key, value):
        self.headers[key] = value

    def set_token(self, token):
        self.headers['Authorization'] = 'Token %s' % token

    def build_url(self, path):
        return urljoin(self.host, path)

    @remote_service('/ping/', json=False)
    def ping(self, **kwargs):
        return self.session.get()

    @remote_service('/notifications/setWebhook/')
    def set_webhook(self, url, token=None):
        """Set webhook used by rebotics_sdk.hooks and rebotics_sdk.notification apps"""
        data = {
            'url': url
        }
        if token is not None:
            data['auth_token'] = token
        return self.session.post(data=data)

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, pformat({
            'host': self.host,
            'headers': self.headers,
            'role': self.role
        }))

    def set_retailer_identifier(self, retailer_id, retailer_secret_key):
        self.headers['x-retailer-id'] = retailer_id
        self.headers['x-retailer-secret-key'] = retailer_secret_key

    def is_retailer_identifier_used(self):
        header_keys = self.headers.keys()
        return 'x-retailer-id' in header_keys and 'x-retailer-secret-key' in header_keys


class PageResult(object):
    def __init__(self, json_data):
        """
        Returns paged result from rest framework
        :param dict json_data:
        """
        self.json_data = json_data

    def __len__(self):
        return self.json_data['count']

    def __iter__(self):
        for item in self.json_data['results']:
            yield item

    def __getitem__(self, item):
        return self.json_data['results'][item]

    @property
    def pdf_download_url(self):
        return self.json_data.get('pdf_url')


def required_model_params(params=None):
    if params is None:
        params = ['model_path', 'index_path', 'meta_path']

    def outer(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            model_params = {
                key: kwargs[key] for key in params
            }
            kwargs['model_params'] = {
                'model_path': model_params['model_path'],
                'model_index_path': model_params['index_path'],
                'model_meta_path': model_params['meta_path']
            }
            result = func(*args, **kwargs)
            return result

        return wrapper

    if callable(params):
        return outer(params)

    return outer



import unittest

from rebotics_sdk.providers.dataset import DatasetProvider


class RetailerProviderTestCase(unittest.TestCase):
    def setUp(self):
        self.dp = DatasetProvider('http://alpha.rebotics.net/')

    def test_api_root(self):
        res = self.dp.api_root()
        assert isinstance(res, dict)



from six import StringIO

from .base import ReboticsBaseProvider, remote_service, required_model_params


class FacenetProvider(ReboticsBaseProvider):
    timeout = 180
    retries = 2

    @required_model_params()
    @remote_service('/images_feature_vectors/')
    def extract_from_image(self, images_io, model_params, **kwargs):
        """Extraction of the feature vectors from image IO"""
        # Expects to have .npy (numpy array) load from cv2.load
        return self.session.post(data=model_params, files=[
            ('images', ('image%d.png' % i, StringIO(image), 'image/png'))
            for i, image in enumerate(images_io)
        ])

    @required_model_params()
    @remote_service('/image_url_feature_vector/')
    def extract_from_image_url(self, images_url, model_params, **kwargs):
        data = model_params
        data['image_link'] = images_url
        return self.session.post(data=data)

    @required_model_params()
    @remote_service('/keyframe_feature_vectors/')
    def extract_from_keyframe(self, keyframe_url, bboxes, model_params, **kwargs):
        data = model_params
        data['bboxes'] = bboxes
        data['keyframe_url'] = keyframe_url
        return self.session.post(data=data)

    @required_model_params()
    @remote_service('/keyframe_image_feature_vector/')
    def extract_from_keyframe_image(self, keyframe_io, bboxes, use_fixed_image_standardization, model_params, **kwargs):
        # Expects to have .npy (numpy array) load from cv2.load
        data = model_params
        data['bboxes'] = bboxes
        data['use_fixed_image_standardization'] = use_fixed_image_standardization
        return self.session.post(
            data=data, files={
                'images': StringIO(keyframe_io)
            }
        )



from .base import ReboticsBaseProvider, remote_service


class ShelfCameraManagerProvider(ReboticsBaseProvider):
    @remote_service('')
    def send_new_file(self):
        self.session.post(

        )

    @remote_service('/api/token-auth/')
    def token_auth(self, username, password):
        json_data = self.session.post(data={
            'username': username,
            'password': password
        })
        self.set_token(json_data['token'])
        return json_data

    @remote_service('/api/folders/')
    def get_root_folders(self):
        return self.session.get()

    @remote_service('/api/camera-settings/')
    def camera_settings_options(self):
        return self.session.options()

    @remote_service('/api/camera-settings/')
    def camera_settings_list(self):
        return self.session.get()

    @remote_service('/api/camera-settings/')
    def add_camera_settings(self, retailer_codename, store_id, token, camera_type, folder, rotation=0):
        """
        Create camera settings entry.
        :param str retailer_codename:
        :param int store_id:
        :param str token:
        :param int camera_type:
        :param str folder:
        :return:
        """
        options = self.camera_settings_options()
        camera_type_choices = options['actions']['POST']['camera_type']['choices']
        assert camera_type in [i['value'] for i in camera_type_choices], \
            'You need to set camera_type based on choices {}'.format(
                [(i['value'], i['display_name']) for i in camera_type_choices]
            )

        return self.session.post(data={
            'retailer': retailer_codename,
            'store_id': store_id,
            'token': token,
            'camera_type': camera_type,
            'camera_folder': folder,
            'rotation': rotation,
        })

    @remote_service('/api/camera-settings/{camera_settings_id}/')
    def camera_settings_detail(self, camera_settings_id):
        return self.session.get(camera_settings_id=camera_settings_id)

    @remote_service('/api/camera-settings/{camera_settings_id}/add_roi/')
    def add_camera_roi(self, camera_settings_id,
                       aisle, section=None,
                       store_planogram_id=None, coordinates=None):
        data = {
            'aisle': aisle,
        }
        if section is not None:
            data['section'] = section

        if store_planogram_id is not None:
            data['store_planogram_id'] = store_planogram_id

        if coordinates is not None:
            assert len(coordinates) == 4, 'Coordinates should iterable and in format [x_min, y_min, x_max, y_max]'
            x_min, y_min, x_max, y_max = coordinates
            data.update({
                'x_min': x_min,
                'y_min': y_min,
                'x_max': x_max,
                'y_max': y_max,
            })

        return self.session.post(camera_settings_id=camera_settings_id, data=data)



import unittest

from rebotics_sdk.providers.retailer import RetailerProvider


class RetailerProviderTestCase(unittest.TestCase):
    def setUp(self):
        self.rp = RetailerProvider('http://alpha.rebotics.net/')

    def test_version(self):
        res = self.rp.version()
        print(res)

    def test_api_v4_root(self):
        res = self.rp.api_v4_root()
        print(res)



from collections import OrderedDict

import requests
import six
from tqdm import tqdm
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor


class ProgressBar(tqdm):

    def update_to(self, n):
        """
        identical to update, except `n` should be current value and not delta.
        """
        self.update(n - self.n)


def upload(destination, file, progress_bar=False):
    url = destination["url"]
    fields = OrderedDict()

    for key in ("policy", "AWSAccessKeyId", "key", "signature"):
        fields[key] = destination["fields"][key]

    fields["file"] = ("features_backup.rcdb", file)
    encoder = MultipartEncoder(fields=fields)
    if not progress_bar:
        return requests.post(
            url, data=encoder, headers={"Content-Type": encoder.content_type}
        )

    with ProgressBar(total=encoder.len, unit="bytes", unit_scale=True, leave=False) as bar:
        monitor = MultipartEncoderMonitor(encoder, lambda monitor: bar.update_to(monitor.bytes_read))
        return requests.post(
            url, data=monitor, headers={"Content-Type": encoder.content_type}
        )


def download(source, destination=None):
    is_file = False
    if destination is None:
        fp = six.BytesIO()
    elif isinstance(destination, six.BytesIO) or isinstance(destination, six.StringIO):
        fp = destination
    else:
        fp = open(destination, 'wb')
        is_file = True

    r = requests.get(source, stream=True)
    file_size = int(r.headers['Content-Length'])
    chunk, chunk_size = 1, 1024
    with ProgressBar(total=file_size, unit='bytes', unit_scale=True, leave=False) as bar:
        for chunk in ProgressBar(
            r.iter_content(chunk_size)
        ):
            fp.write(chunk)

    if is_file:
        fp.close()
        return destination
    return fp



from .base import *  # NOQA
from .admin import AdminProvider  # NOQA
from .dataset import DatasetProvider  # NOQA
from .facenet import FacenetProvider  # NOQA
from .retailer import RetailerProvider  # NOQA



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `rebotics_sdk` package."""

import pytest
from click.testing import CliRunner



@pytest.fixture
def response():
    """Sample pytest fixture.

    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')


def test_content(response):
    """Sample pytest test function with the pytest fixture as an argument."""
    # from bs4 import BeautifulSoup
    # assert 'GitHub' in BeautifulSoup(response.content).title.string


def test_roles_invoke():
    runner = CliRunner()
    from rebotics_sdk.cli.retailer import api

    result = runner.invoke(api, ['roles'])
    assert result.exit_code == 0


def test_processing_action_group_invoke():
    runner = CliRunner()
    from rebotics_sdk.cli.retailer import api

    result = runner.invoke(api, ['processing-action'])

    result = runner.invoke(api, ['processing-action', 'download'])


def test_configure_command():
    runner = CliRunner()
    # result = runner.invoke(api, ['-r', 'local', 'configure'])


def test_camera_settings_add_roi():
    runner = CliRunner()
    from rebotics_sdk.cli.shelf_camera_manager import api

    result = runner.invoke(api, ['camera-settings', 'add-roi'])



import datetime
import json
import os
import pathlib
from functools import partial
from multiprocessing.pool import Pool

import click
import humanize
from prettytable import PrettyTable

from rebotics_sdk.advanced import remote_loaders
from rebotics_sdk.cli.common import shell, configure, roles
from .utils import ReboticsCLIContext, app_dir, pass_rebotics_context, read_saved_role, process_role
from ..advanced.packers import ClassificationDatabasePacker, ClassificationDatabaseException
from ..providers import AdminProvider, RetailerProvider, sleep
from ..providers.facenet import FacenetProvider
from ..utils import Timer


@click.group()
@click.option('-f', '--format', default='table', type=click.Choice(['table', 'id', 'json']), help='Result rendering')
@click.option('-v', '--verbose', is_flag=True, help='Enables verbose mode')
@click.option('-c', '--config', type=click.Path(), default='admin.json', help="Specify what config.json to use")
@click.option('-r', '--role', default=lambda: read_saved_role('admin'), help="Key to specify what admin to use")
@click.version_option()
@click.pass_context
def api(ctx, format, verbose, config, role):
    """
    Admin CLI tool to communicate with dataset API
    """
    process_role(ctx, role, 'admin')
    ctx.obj = ReboticsCLIContext(
        role,
        format,
        verbose,
        os.path.join(app_dir, config),
        provider_class=AdminProvider
    )


def get_retailer_version_task(retailer_dict):
    retailer_provider = RetailerProvider(host=retailer_dict['host'], retries=1, timeout=5)
    try:
        response = retailer_provider.version()
        version = response['version']
        uptime = humanize.naturaldelta(datetime.timedelta(seconds=int(response['uptime'])))
    except Exception:
        version = 'not working'
        uptime = '---'

    d = [
        retailer_dict['codename'],
        retailer_dict['title'],
        version,
        uptime,
        retailer_dict['host'],
    ]
    return d


@api.command()
@click.argument('retailer')
@pass_rebotics_context
def configurations(ctx, retailer):
    try:
        res = ctx.provider.get_configurations(retailer)
        ctx.format_result(res)
    except Exception as exc:
        raise click.ClickException(exc)


@api.command()
@click.option('-n', '--notify', is_flag=True)
@click.option('-d', '--delay', type=click.INT, default=60)
@pass_rebotics_context
def retailer_versions(ctx, notify, delay):
    """Fetch retailer versions and their meta information"""
    if notify:
        if ctx.verbose:
            click.echo('Using notify option', err=True)

        try:
            from pynotifier import Notification
            Notification(
                title='Subscribed to the notifications',
                description='You will receive notifications for retailer updates',
            ).send()
        except ImportError:
            raise click.ClickException("You can't use notify function")

    provider = ctx.provider
    if ctx.verbose:
        click.echo('Fetching info from rebotics admin.', err=True)
    retailers = provider.get_retailer_list()
    prev_results = []
    results = []
    pool = Pool(len(retailers))

    while True:
        try:
            if ctx.verbose:
                click.echo('Fetching the retailer versions', err=True)
            results = pool.map(get_retailer_version_task, retailers)

            if not notify:
                break

            for prev_result in prev_results:
                retailer_codename = prev_result[0]
                previous_version = prev_result[2]
                for result in results:
                    if result[0] == retailer_codename:
                        current_version = result[2]
                        if previous_version != current_version:
                            notification_message = 'Retailer {} updated from version {} to {}'.format(
                                retailer_codename,
                                previous_version,
                                current_version
                            )
                            click.echo(notification_message)
                            Notification(
                                title=notification_message,
                                description='Current uptime is: {}'.format(result[3]),
                                duration=30,
                                urgency=Notification.URGENCY_CRITICAL,
                            ).send()
            del prev_results
            prev_results = results
            sleep(delay)
        except KeyboardInterrupt:
            break

    table = PrettyTable()
    table.field_names = ['codename', 'title', 'version', 'uptime', 'host']
    for result in results:
        table.add_row(result)
    click.echo(table)


def load_models(ctx, retailer_id, retailer_secret):
    provider = ctx.provider
    provider.set_retailer_identifier(retailer_id, retailer_secret)
    return provider.get_retailer_tf_models()


@api.command()
@click.option('-r', '--retailer-id', help='Retailer id')
@click.option('-s', '--retailer-secret', help='Retailer secret key')
@click.option('-u', '--facenet-url', help='Facenet service URL')
@click.argument('image_url')
@pass_rebotics_context
def extract_feature_vectors(ctx, retailer_id, retailer_secret, facenet_url, image_url):
    """Fetches latest configuration of neural model for retailer by it's ID and Secret key;
    Sends image to facenet to load model into state."""
    models = load_models(ctx, retailer_id, retailer_secret)

    facenet_model = models['facenet_model']
    if ctx.verbose:
        click.echo("Facenet model: %s" % facenet_model['codename'])

    facenet_provider = FacenetProvider(facenet_url)
    feature_extractor = partial(
        facenet_provider.extract_from_image_url,
        model_path=facenet_model['data_path'],
        index_path=facenet_model['index_path'],
        meta_path=facenet_model['meta_path'],
    )

    with Timer() as t:
        result = feature_extractor(image_url)
        click.echo(result)

    if ctx.verbose:
        click.echo("Elapsed: %s seconds" % t.elapsed_secs)


@api.command()
@click.option('-r', '--retailer-id', help='Retailer id')
@click.option('-s', '--retailer-secret', help='Retailer secret key')
@click.option('-u', '--facenet-url', help='Facenet service URL')
@click.argument('image_url')
@click.argument('bounding_boxes', type=click.File())
@pass_rebotics_context
def extract_feature_vectors_for_boxes(ctx, retailer_id, retailer_secret, facenet_url, image_url, bounding_boxes):
    """Fetches latest configuration of neural model for retailer by it's ID and Secret key;
    Sends keyframe image url and list of bounding boxes to facenet to load model into state."""
    models = load_models(ctx, retailer_id, retailer_secret)

    facenet_model = models['facenet_model']
    if ctx.verbose:
        click.echo("Facenet model: %s" % facenet_model['codename'])

    boxes = json.load(bounding_boxes)
    assert isinstance(boxes, list), "Need to supply list of bounding boxes"

    facenet_provider = FacenetProvider(facenet_url)
    feature_extractor = partial(
        facenet_provider.extract_from_keyframe,
        model_path=facenet_model['data_path'],
        index_path=facenet_model['index_path'],
        meta_path=facenet_model['meta_path'],
    )

    with Timer() as t:
        result = feature_extractor(keyframe_url=image_url, bboxes=boxes)
        click.echo(result)

    if ctx.verbose:
        click.echo("Elapsed: %s seconds" % t.elapsed_secs)


@api.command()
@click.argument('retailer', type=click.STRING)
@click.argument('url', type=click.STRING)
@pass_rebotics_context
def set_retailer_url(ctx, retailer, url):
    try:
        ctx.provider.update_host(retailer, url)
    except Exception as exc:
        raise click.ClickException(str(exc))
    else:
        click.echo('Set new host for retailer %s' % retailer)


def upload_packed(ctx, import_request, packed):
    destination = import_request['destination']

    if ctx.verbose:
        click.echo("Sending the file to the S3 {}".format(destination))

    response = remote_loaders.upload(destination, packed, progress_bar=True)

    if response.status_code < 300:
        if ctx.verbose:
            click.echo("S3 {} response: {}".format(response.status_code, response.content))

        ctx.provider.notify_classification_database_import_done(
            id=import_request['id']
        )
        click.echo('File was uploaded to destination. Classification data backup ID={}'.format(import_request['id']))
    else:
        raise click.ClickException("Failed to upload classification file. Destination status: {}".format(
            response.status_code)
        )


@api.command()
@click.option('-r', '--retailer', help="Retailer codename", prompt=True)
@click.option('-m', '--model', help="Model codename", prompt=True)
@click.option('-i', '--images', type=click.Path(exists=True, file_okay=False, dir_okay=True))
@click.option('-l', '--labels', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.option('-f', '--features', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@pass_rebotics_context
def pack_and_import_classification_db(ctx, retailer, model, images, labels, features):
    """Pack classification database to single .rcdb file and import it to the Rebotics Admin

\b
db/
├── custom_folder
│   ├── image_2.png
│   └── image_1.png
├── features.txt
└── labels.txt

    It is a single step command with equivalent of running two commands:

admin pack-classification-db --features features.txt --labels labels.txt --images ./custom_folder/ --target classification.rcdb

admin import_classification_db --retailer delta --model test_code classification.rcdb
    """
    packer = ClassificationDatabasePacker(destination=None, progress_bar=True)
    if ctx.verbose:
        click.echo("Packing from provided values: \n"
                   "labels: {labels} \n"
                   "features: {features} \n"
                   "images: {images}".format(labels=labels, features=features, images=images))
    try:
        packed = packer.pack(labels, features, images)
        packed.seek(0)
    except ClassificationDatabaseException as exc:
        raise click.ClickException(str(exc))

    if ctx.verbose:
        click.echo("Creating import request: \n"
                   "retailer: {}\n"
                   "model: {}\n"
                   "extension: {}".format(retailer, model, packer.extension))

    import_request = ctx.provider.create_classification_database_import(
        retailer=retailer, model=model, extension=packer.extension
    )
    if ctx.verbose:
        click.echo("Import request: {}".format(import_request))

    upload_packed(ctx, import_request, packed)


@api.command()
@click.option('-r', '--retailer', help="Retailer codename", prompt=True)
@click.option('-m', '--model', help="Model codename", prompt=True)
@click.argument('filepath', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@pass_rebotics_context
def import_classification_db(ctx, retailer, model, filepath):
    """Import rcdb file to Rebotics Admin. Example usage:

    admin import_classification_db --retailer delta --model test_code classification.rcdb

    """
    extension = os.path.split(filepath)[-1].split('.')[-1]
    if ctx.verbose:
        click.echo("Creating import request: \n"
                   "retailer: {}\n"
                   "model: {}\n"
                   "extension: {}".format(retailer, model, extension))

    import_request = ctx.provider.create_classification_database_import(
        retailer=retailer, model=model, extension=extension
    )
    if ctx.verbose:
        click.echo("Import request: {}".format(import_request))
    with open(filepath, 'rb') as packed:
        upload_packed(ctx, import_request, packed)


@api.command()
@click.option('-i', '--images', type=click.Path(exists=True, file_okay=False, dir_okay=True))
@click.option('-l', '--labels', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.option('-f', '--features', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.argument('target', type=click.Path(exists=False, file_okay=True, dir_okay=True), default='.')
@pass_rebotics_context
def pack_classification_db(ctx, images, labels, features, target):
    """Pack classification database to single .rcdb file

\b
db/
├── custom_folder
│   ├── image_2.png
│   └── image_1.png
├── features.txt
└── labels.txt

    Example usage:
        admin pack-classification-db --features features.txt --labels labels.txt --images ./custom_folder/ --target classification.rcdb

    """
    target = pathlib.Path(target)
    if target.is_dir():
        target = target / "classification_db{}.{}".format(
            datetime.datetime.now().strftime('%Y-%m-%dZ%H%M'),
            ClassificationDatabasePacker.extension
        )
    packer = ClassificationDatabasePacker(destination=target, progress_bar=True)
    if ctx.verbose:
        click.echo("Packing from provided values: \n"
                   "labels: {labels} \n"
                   "features: {features} \n"
                   "images: {images}".format(labels=labels, features=features, images=images))
    try:
        packed = packer.pack(labels, features, images)
    except ClassificationDatabaseException as exc:
        raise click.ClickException(str(exc))
    click.echo('Written to {}'.format(packed))


api.add_command(shell, 'shell')
api.add_command(roles, 'roles')
api.add_command(configure, 'configure')



import django.dispatch

rebotics_update = django.dispatch.Signal(
    providing_args=['retailer', 'component', 'status', 'body']
)
processing_succeed = django.dispatch.Signal(
    providing_args=['retailer', 'processing_id', 'body']
)
processing_started = django.dispatch.Signal(
    providing_args=['retailer', 'processing_id', 'body']
)
processing_failed = django.dispatch.Signal(
    providing_args=['retailer', 'processing_id', 'body']
)

processing_deleted = django.dispatch.Signal(
    providing_args=['retailer', 'processing_id', 'body']
)
processing_update_received = django.dispatch.Signal(
    providing_args=['retailer', 'processing_id', 'status', 'body']
)


class SignalEmitter(object):
    def __init__(self, retailer, component, status, body, request):
        self.retailer = retailer
        self.body = body
        self.status = status
        self.component = component
        self.request = request

    def emit(self):
        rebotics_update.send_robust(
            sender=self,
            retailer=self.retailer,
            component=self.component,
            status=self.status,
            body=self.body
        )


class ProcessingSignalEmitter(SignalEmitter):
    started = processing_started
    succeed = processing_succeed
    failed = processing_failed
    received = processing_update_received
    deleted = processing_deleted

    def emit(self):
        super(ProcessingSignalEmitter, self).emit()
        processing_id = self.body['id']

        self.received.send_robust(
            sender=self,
            processing_id=processing_id,
            status=self.status,
            retailer=self.retailer,
            body=self.body
        )

        getattr(self, self.status).send_robust(
            sender=self,
            retailer=self.retailer,
            processing_id=processing_id,
            body=self.body
        )


def emitter_factory(retailer, component, status, body, request):
    components = {
        'processing': ProcessingSignalEmitter,
        'internal': SignalEmitter,
    }
    emitter = components.get(component, SignalEmitter)

    return emitter(retailer, component, status, body, request)



from rest_framework import status
from rest_framework.response import Response
from rest_framework.views import APIView

from rebotics_sdk.hook import signals
from rebotics_sdk.hook.serializers import WebhookDataSerializer


class WebhookHandlerViewSet(APIView):
    permission_classes = []

    def post(self, request, *args, **kwargs):
        retailer_code = kwargs['retailer_code']
        serializer = WebhookDataSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        emitter = signals.emitter_factory(
            retailer_code,
            serializer.validated_data['component'],
            serializer.validated_data['status'],
            serializer.validated_data['body'],
            request)
        emitter.emit()

        return Response(status=status.HTTP_202_ACCEPTED)



import json
from datetime import datetime

import click
import pytz
from dateutil import parser as dp
from prettytable import PrettyTable


def extract_fields_to_render(d, max_column_length, keys_to_skip, key_prefix=None, depth=0, max_depth=1):
    fields_to_render = []

    if depth > max_depth:
        return []

    for field, value in d.items():
        if field in keys_to_skip:
            continue

        if key_prefix is not None:
            field = '{}.{}'.format(key_prefix, field)

        if not (isinstance(value, list)):
            if isinstance(value, str):
                if len(value) < max_column_length:
                    fields_to_render.append(field)
            elif isinstance(value, int) or isinstance(value, float):
                fields_to_render.append(field)
            elif value is None:
                fields_to_render.append(field)
            elif isinstance(value, dict):
                fields_to_render.extend(
                    extract_fields_to_render(value, max_column_length, keys_to_skip, key_prefix=field)
                )
    return fields_to_render


def format_full_table(results, max_column_length=30, keys_to_skip=None, fields=None):
    if keys_to_skip is None:
        keys_to_skip = []
    if fields is None:
        fields = []

    if fields and keys_to_skip:
        raise ValueError('Can only use either keys_to_skip or either fields, not both.')

    def get_attribute(value, key, default):
        if '.' not in key:
            return value.get(key, default)

        keys = key.split('.')
        acc = value
        for k in keys:
            try:
                acc = acc.get(k, '---')
            except KeyError:
                acc = default
        return acc

    if not isinstance(results, list):
        results = [results]
    table = PrettyTable()
    for i, item in enumerate(results):
        if i == 0:
            if fields:
                table.field_names = fields
            else:
                table.field_names = extract_fields_to_render(item, max_column_length, keys_to_skip)

        table.add_row([
            get_attribute(item, field, '---') for field in table.field_names
        ])
    click.echo(table)


def format_processing_action_output(processing_actions_list, format_):
    if format_ == 'json':
        click.echo(json.dumps(processing_actions_list, indent=2))
    elif format_ == 'id':
        click.echo(" ".join([str(item['id']) for item in processing_actions_list]))
    else:
        pac = processing_actions_list[0]

        if 'last_requeue' not in pac.keys():
            fields = ['id', 'store', 'username', 'status', 'store_planogram_id', 'aisle', 'section', 'created']
        else:
            fields = ['id', 'store', 'username', 'status', 'store_planogram_id', 'aisle', 'section', 'created',
                      'last_requeue', 'time_in_queue']

        format_full_table([PacSerializer(x) for x in processing_actions_list], fields=fields)


class PacSerializer(object):
    def __init__(self, data):
        self.data = data
        self.now = datetime.now(pytz.utc)

    def get(self, item, default='---'):
        return self.data.get(item) or getattr(self, item, default)

    @property
    def time_in_queue(self):
        if self.data['status'] not in ['error', 'done', 'interrupted']:
            last_requeue_datetime = self.last_requeue
            if last_requeue_datetime:
                return self.now - last_requeue_datetime

        processing_start_time = self.data.get('processing_start_time')
        processing_finish_time = self.data.get('processing_finish_time')
        if processing_finish_time is None and processing_finish_time is None:
            time_in_queue = 'unknown'
        else:
            start_time = dp.parse(processing_start_time)
            finish_time = dp.parse(processing_finish_time)
            time_in_queue = finish_time - start_time

        return time_in_queue

    @property
    def created(self):
        return dp.parse(self.data['created']).strftime('%c')

    @property
    def username(self):
        u = self.data.get('user')
        if u:
            return u.get('username', 'id')

    @property
    def last_requeue(self):
        d = self.data.get('last_requeue')
        if d is None:
            return None

        return dp.parse(d)

    @property
    def store(self):
        return '#{}'.format(self.data.get('store_id'))



import json
import logging
import pathlib
import re
import zipfile
from collections import namedtuple
from datetime import datetime
from io import BytesIO, StringIO

import rebotics_sdk
import tqdm

logger = logging.getLogger(__name__)

ClassificationEntry = namedtuple('ClassificationEntry', ['label', 'feature', 'image', 'filename'])
ImageEntry = namedtuple('ImageEntry', ['filename', 'filepath', 'order'])

NUMERIC_VALUE = re.compile(r'\d+')


class ClassificationDatabaseException(Exception):
    pass


def extract_numeric(filename):
    found = NUMERIC_VALUE.findall(filename)
    if len(found) == 0:
        raise ClassificationDatabaseException('Images name should contain numeric value that represents the order. '
                                              'Instead of {} name was used'.format(filename))
    return int(found[0])


class BaseClassificationDatabasePacker(object):
    version = 0
    extension = 'zip'

    def __init__(self, source=None, destination=None, progress_bar=False):
        self.source = source
        self.destination = destination
        self.meta_data = {
            'packed': datetime.now().strftime('%c'),
            'sdk_version': rebotics_sdk.__version__,
            'packer_version': self.version
        }
        self.images = []

        if self.destination is None:
            self.destination = BytesIO()
        self.progress_bar = progress_bar

    def read_lines(self, lines):
        return [x for x in lines.split('\n') if x]

    def zipfile(self, file, **kwargs):
        # if python version is
        params = dict(
            compression=zipfile.ZIP_DEFLATED,
            allowZip64=True,
        )
        #
        # if (sys.version_info.major, sys.version_info.minor) >= (3, 7):
        #     # using the most aggressive compression
        #     params['compresslevel'] = 9  # default is 6

        params.update(kwargs)

        for compression_option in [zipfile.ZIP_DEFLATED, zipfile.ZIP_STORED]:
            try:
                params['compression'] = compression_option
                return zipfile.ZipFile(str(file), **params)
            except RuntimeError:
                pass

    def pack(self, *args, **kwargs):
        raise NotImplementedError()

    def unpack(self):
        raise NotImplementedError()


class ZipDatabasePacker(BaseClassificationDatabasePacker):
    """
    Unified file format for packing classification database into single folder
    File format is the zip based archive with following :

    labels.txt
    features.txt
    meta.json
    """
    version = 0
    extension = 'zip'

    def pack(self, labels, features, *args, **kwargs):
        labels_io = StringIO("\n".join(labels))
        features_io = StringIO("\n".join(features))
        self.meta_data['count'] = len(labels)

        with self.zipfile(self.destination, mode='w') as zip_io:
            zip_io.writestr('labels.txt', labels_io.getvalue())
            zip_io.writestr('features.txt', features_io.getvalue())
            zip_io.writestr('meta.json', json.dumps(self.meta_data))
        return self.destination

    def unpack(self):
        if self.source is None:
            raise ClassificationDatabaseException("Can not unpack with empty source file")

        with self.zipfile(self.source, mode='r') as zip_io:
            try:
                self.meta_data = json.load(zip_io.open('meta.json'))
            except KeyError:
                # meta.json is not presented, working in the compatibility mode
                pass

            labels = self.read_lines(zip_io.read('labels.txt').decode('utf-8'))
            features = self.read_lines(zip_io.read('features.txt').decode('utf-8'))

            for label, feature in zip(
                labels,
                features,
            ):
                yield ClassificationEntry(
                    label, feature, None, None
                )


class ClassificationDatabasePacker(BaseClassificationDatabasePacker):
    """
    Unified file format for packing classification database into single folder
    File format is the zip based archive with following :

    labels.txt
    features.txt
    meta.json
    images/  - folder with images, this folder can be empty
    """

    version = 1
    extension = 'rcdb'
    images_extensions = {'jpeg', 'png', 'jpg'}

    def extract_meta_data(self, labels_path, features_path, images_folder):
        known_order = set()
        for f in pathlib.Path(images_folder).iterdir():
            if f.is_file():
                extension = f.suffix
                if extension in self.images_extensions:
                    raise ClassificationDatabaseException("Extension is not supported by packer: {}".format(f.name))
                numeric_order = extract_numeric(f.name)
                if numeric_order in known_order:
                    raise ClassificationDatabaseException("Numeric order collision. {}".format(f.name))

                known_order.add(numeric_order)
                self.images.append(ImageEntry(f.name, str(f), numeric_order))

        # repack based on the filename and it's numeric value
        self.images.sort(key=lambda i: i.order)  # sort by numeric value of the filename

        with open(labels_path, 'r') as labels_io, open(features_path, 'r') as features_io:
            labels = self.read_lines(labels_io.read())
            features = self.read_lines(features_io.read())

        labels_count = len(labels)
        features_count = len(features)
        files_count = len(self.images)

        if labels_count != features_count:
            raise ClassificationDatabaseException('Inconsistent count of labels and features. {}/{}'.format(
                labels_count, features_count
            ))

        if labels_count != files_count:
            raise ClassificationDatabaseException(
                'Inconsistent count of labels and features and files. {}/{}/{}'.format(
                    labels_count, features_count, files_count
                ))

        self.meta_data.update({
            'count': features_count,
            'images': [str(img.filename) for img in self.images]
        })

    def pack(self, labels_path, features_path, images_folder):
        self.extract_meta_data(labels_path, features_path, images_folder)

        with self.zipfile(self.destination, mode='w') as zip_io:
            zip_io.write(labels_path, 'labels.txt')
            zip_io.write(features_path, 'features.txt')
            zip_io.writestr('meta.json', json.dumps(self.meta_data))

            if not self.progress_bar:
                for img in self.images:
                    zip_io.write(img.filepath, str(pathlib.Path('images') / img.filename))
            else:
                for img in tqdm.tqdm(self.images):
                    zip_io.write(img.filepath, str(pathlib.Path('images') / img.filename))

        return self.destination

    def unpack(self):
        """

        :return: generator of label, feature, image_io
        :rtype:self.meta_data
        """
        if self.source is None:
            raise ClassificationDatabaseException("Can not unpack with empty source file")

        with self.zipfile(self.source, mode='r') as zip_io:
            self.meta_data = json.load(zip_io.open('meta.json'))

            labels = self.read_lines(zip_io.read('labels.txt').decode('utf-8'))
            features = self.read_lines(zip_io.read('features.txt').decode('utf-8'))

            for label, feature, image_name in zip(
                labels,
                features,
                self.meta_data['images']
            ):
                image = zip_io.read('images/{}'.format(image_name))
                yield ClassificationEntry(
                    label, feature, image, image_name
                )



import errno
import os
import re
from timeit import default_timer

import requests
from six.moves.urllib_parse import urlparse


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def get_filename_from_url(url):
    return urlparse(url).path.split('/')[-1]


def download_file(url, filepath=None, **params):
    response = requests.get(url, stream=True, params=params)
    response.raise_for_status()

    if filepath is None:
        # TODO: decode url, get filename
        filepath = get_filename_from_url(url)

    with open(filepath, 'wb') as handle:
        for block in response.iter_content(1024):
            handle.write(block)
    return filepath


class Timer(object):
    def __init__(self):
        self.timer = default_timer

    def __enter__(self):
        self.start = self.timer()
        return self

    def __exit__(self, *args):
        end = self.timer()
        self.elapsed_secs = end - self.start
        self.elapsed = self.elapsed_secs * 1000  # millisecs


def parse_id_range_string(id_string):
    id_list = []
    range_pattern = re.compile(r'^(?P<from>\d+)-(?P<to>\d+)$')
    single_pattern = re.compile(r'^\d+$')

    left_out_parts = []

    for part in id_string.split(','):
        part = part.strip()
        range_match = range_pattern.match(part)

        if range_match:
            range_from = int(range_match.group('from'))
            range_to = int(range_match.group('to'))
            if range_from > range_to:
                left_out_parts.append(part)
                continue
            id_list.extend(range(range_from, range_to + 1))
        elif single_pattern.match(part):
            id_list.append(int(part))
        else:
            left_out_parts.append(part)

    if left_out_parts:
        raise ValueError('Please review the string as input: {}.\n Wrong parts are: {}'.format(
            id_string, left_out_parts
        ))
    if not id_list:
        raise ValueError('You need to supply a string with at least one ID. Given {}'.format(id_string))
    return sorted(list(set(id_list)))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-05-27 09:38
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('notifications', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Message',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('modified', models.DateTimeField(auto_now=True, verbose_name='Date Modified')),
                ('component', models.CharField(choices=[(b'processing', 'Processing action'), (b'internal', 'Internal')], max_length=64)),
                ('status', models.CharField(max_length=32)),
                ('body', models.TextField()),
                ('delivery_status', models.CharField(choices=[(b'message_pending', 'Pending'), (b'message_delivered', 'Delivered'), (b'message_rejected', 'Rejected'), (b'message_not_delivered', 'Not delivered')], max_length=64)),
                ('exception_message', models.TextField()),
            ],
            options={
                'verbose_name': 'Message',
                'verbose_name_plural': 'Messages',
            },
        ),
        migrations.AlterField(
            model_name='webhookrouter',
            name='auth_token',
            field=models.CharField(max_length=64, null=True),
        ),
        migrations.AddField(
            model_name='message',
            name='receiver',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='messages', to='notifications.WebhookRouter'),
        ),
    ]



import pathlib

from rebotics_sdk.advanced.packers import ClassificationDatabasePacker, ZipDatabasePacker


def test_classification_packing():
    packer = ClassificationDatabasePacker(destination='test')
    features = pathlib.Path('db/features.txt')
    labels = pathlib.Path('db/labels.txt')
    images_folder = pathlib.Path('db/custom_folder/')

    res = packer.pack(labels, features, images_folder)

    assert res == 'test.rcdb'
    assert len(packer.images) == 2

    packer = ClassificationDatabasePacker(source='test.rcdb')
    entries = list(packer.unpack())
    assert len(entries) == 2
    entry = entries[0]
    assert entry.label == '123123123'
    assert entry.feature == '123123123123123'
    internal_filename = entry.filename
    assert internal_filename == 'image_1.png'

    # testing if it can be dumped to the FS
    og_file = pathlib.Path('db/custom_folder') / internal_filename
    tmp_file = pathlib.Path('db') / internal_filename

    with open(tmp_file, 'wb') as fout:
        fout.write(entry.image)

    assert og_file.stat().st_size == tmp_file.stat().st_size
    tmp_file.unlink()


def test_zip_packing():
    packer = ZipDatabasePacker()
    packed = packer.pack(
        labels=[
            '123123123'
        ],
        features=[
            '123123123123123'
        ]
    )
    assert packer.meta_data['count'] == 1

    unpacker = ZipDatabasePacker(source=packed)
    for entry in unpacker.unpack():
        assert entry.label == '123123123'
        assert entry.feature == '123123123123123'

    assert unpacker.meta_data['count'] == 1
